import pandas as pd
import networkx as nx
import qiime2

from .plugin_setup import plugin
from ._format import PairwiseFeatureDataFormat, GraphModelingLanguageFormat, ModuleMembershipTSVFormat


@plugin.register_transformer
def _1(data: pd.DataFrame) -> PairwiseFeatureDataFormat:
    ff = PairwiseFeatureDataFormat()
    with ff.open() as fh:
        data.to_csv(fh, sep='\t', index_label=('feature1', 'feature2'))
    return ff


@plugin.register_transformer
def _2(ff: PairwiseFeatureDataFormat) -> pd.DataFrame:
    df = pd.read_table(str(ff), index_col=(0, 1), sep='\t')
    new_index = pd.MultiIndex.from_tuples([(str(i), str(j)) for i, j in df.index])
    df.index = new_index
    return df


@plugin.register_transformer
def _3(data: nx.Graph) -> GraphModelingLanguageFormat:
    ff = GraphModelingLanguageFormat()
    nx.write_gml(data, str(ff.path))
    return ff


@plugin.register_transformer
def _4(ff: GraphModelingLanguageFormat) -> nx.Graph:
    return nx.read_gml(str(ff))


@plugin.register_transformer
def _5(data: pd.Series) -> ModuleMembershipTSVFormat:
    ff = ModuleMembershipTSVFormat()
    with ff.open() as fh:
        data.to_csv(fh, sep='\t')
    return ff


@plugin.register_transformer
def _6(ff: ModuleMembershipTSVFormat) -> qiime2.Metadata:
    df = pd.read_table(str(ff), index_col=0, header=None, dtype=str)
    df.columns = ['module_membership']
    df.index = [str(i) for i in df.index]
    df.index.name = '#OTU ID'
    return qiime2.Metadata(df)



import qiime2.plugin.model as model
from qiime2.plugin import ValidationError
import networkx as nx
import pandas as pd
from itertools import combinations


class PairwiseFeatureDataFormat(model.TextFileFormat):
    def _validate_(self, level='min'):
        if level == 'min':
            nrows = 8
        elif level == 'max':
            nrows = None
        else:
            raise ValueError('Nonstandard level value')
        try:
            frame = pd.read_table(self.path, index_col=(0, 1), sep='\t', nrows=nrows)
        except IndexError:
            raise ValidationError('Only one column in pairwise file')
        if frame.empty:
            raise ValidationError('No data in file')
        if level == 'max':
            # validate
            all_features = set([feature for feature_pair in frame.index for feature in feature_pair])
            all_feature_pairs = set(frozenset(pair) for pair in combinations(all_features, 2))
            all_found_pairs = [frozenset(pair) for pair in frame.index]
            if len(all_found_pairs) != len(set(all_found_pairs)):
                raise ValidationError('pair is repeated')
            if all_feature_pairs != set(all_found_pairs):
                raise ValidationError('Not all features pairs present in table')


PairwiseFeatureDataDirectoryFormat = model.SingleFileDirectoryFormat(
    'PairwiseFeatureDataDirectoryFormat', 'pairwise_comparisons.tsv', PairwiseFeatureDataFormat)


class GraphModelingLanguageFormat(model.TextFileFormat):
    def _validate_(self, level):
        try:
            _ = nx.read_gml(str(self.path))
        except nx.NetworkXError:
            raise ValidationError('Not a valid GML file')


GraphModelingLanguageDirectoryFormat = model.SingleFileDirectoryFormat(
    'GraphModelingLanguageDirectoryFormat', 'network.gml', GraphModelingLanguageFormat)


class ModuleMembershipTSVFormat(model.TextFileFormat):
    def _validate_(self, level):
        if level == 'min':
            nrows = 8
        elif level == 'max':
            nrows = None
        else:
            raise ValueError('Nonstandard level value')
        series = pd.read_table(self.path, header=None, squeeze=True, nrows=nrows, index_col=0)
        if type(series) != pd.Series:
            raise ValidationError('File has more than one column: %s' % series.head())


ModuleMembershipTSVDirectoryFormat = model.SingleFileDirectoryFormat(
    'ModuleMembershipTSVDirectoryFormat', 'module-membership.tsv', ModuleMembershipTSVFormat)



import importlib

from qiime2.plugin import (Str, Plugin, Choices, Float, Range, Bool, Citations, Int)
from q2_types.feature_table import FeatureTable, Frequency

from ._type import Network, PairwiseFeatureData, ModuleMembership
from ._format import GraphModelingLanguageFormat, GraphModelingLanguageDirectoryFormat, PairwiseFeatureDataFormat, \
                     PairwiseFeatureDataDirectoryFormat, ModuleMembershipTSVFormat, ModuleMembershipTSVDirectoryFormat
from ._SCNIC_methods import sparcc_filter, calculate_correlations, build_correlation_network_r, \
                            build_correlation_network_p, make_modules_on_correlations

import q2_SCNIC

citations = Citations.load('citations.bib', package='q2_SCNIC')

plugin = Plugin(
    name='SCNIC',
    version=q2_SCNIC.__version__,
    website="https://github.com/shafferm/q2-SCNIC",
    package='q2_SCNIC',
    description=(
        'This QIIME 2 plugin allows for use of the SCNIC methods '
        'to build correlation networks as well as detect and '
        'summarize modules of highy intercorrelated features'),
    short_description='Plugin for SCNIC usage.',
    citations=[citations['SciPyProceedings_11']]
)

plugin.register_semantic_types(Network)
plugin.register_semantic_types(PairwiseFeatureData)
plugin.register_semantic_types(ModuleMembership)

plugin.register_formats(GraphModelingLanguageFormat)
plugin.register_formats(GraphModelingLanguageDirectoryFormat)
plugin.register_formats(PairwiseFeatureDataFormat)
plugin.register_formats(PairwiseFeatureDataDirectoryFormat)
plugin.register_formats(ModuleMembershipTSVFormat)
plugin.register_formats(ModuleMembershipTSVDirectoryFormat)

plugin.register_semantic_type_to_format(Network, artifact_format=GraphModelingLanguageDirectoryFormat)
plugin.register_semantic_type_to_format(PairwiseFeatureData, artifact_format=PairwiseFeatureDataDirectoryFormat)
plugin.register_semantic_type_to_format(ModuleMembership, artifact_format=ModuleMembershipTSVDirectoryFormat)

plugin.methods.register_function(
    function=sparcc_filter,
    inputs={'table': FeatureTable[Frequency]},
    parameters={},
    outputs=[('table_filtered', FeatureTable[Frequency])],
    input_descriptions={'table': (
        'Table to be filtered for use in correlational analyses.')},
    output_descriptions={'table_filtered': 'The filtered table.'},
    name='Filter a table based on recommended parameters from Friedman et al. 2012',
    description=(
        'Filter a feature table based on recommendations from Friedman et al. 2012.'
        'Samples with a count of less than 500 and features with mean abundance of '
        '<2 are removed.'
    ),
)

plugin.methods.register_function(
    function=calculate_correlations,
    inputs={'table': FeatureTable[Frequency]},  # TODO: Generalize, don't require frequency
    parameters={'method': Str % Choices(['kendall', 'pearson', 'spearman', 'sparcc']),
                'p_adjustment_method': Str, 'n_procs': Int},
    outputs=[('correlation_table', PairwiseFeatureData)],
    input_descriptions={'table': (
        'Normalized and filtered feature table to use for microbial interdependence test.')},
    parameter_descriptions={
        'method': 'The correlation test to be applied.',
        'p_adjustment_method': 'The method for p-value adjustment to be applied. '
                               'This can be selected from the list of methods in '
                               'statsmodels multipletests.',
        'n_procs': 'Number of processors to use in correlation analysis'
    },
    output_descriptions={'correlation_table': 'The resulting table of pairwise correlations with R and p-value.'},
    name='Build pairwise correlations between observations',
    description=(
        'Build pairwise correlations between all observations in feature table'
    ),
    citations=[citations['Friedman2012'], citations['Watts2018']]
)

plugin.methods.register_function(
    function=build_correlation_network_r,
    inputs={'correlation_table': PairwiseFeatureData},
    parameters={'min_val': Float % Range(0, 1, inclusive_end=True),
                'cooccur': Bool},
    outputs=[('correlation_network', Network)],
    input_descriptions={'correlation_table': (
        'Pairwise feature data table of correlations with r value.')},
    parameter_descriptions={
        'min_val': 'The minimum r value to say an edge should exist.',
        'cooccur': 'Whether or not to constrain the network to only positive edges.'
    },
    output_descriptions={'correlation_network': 'The resulting network.'},
    name='Build a correlation network based on an r value cutoff',
    description=(
        'Build a correlation network where nodes are features and edges '
        'are correlations are stronger than the provided min_val.'
    ),
)

plugin.methods.register_function(
    function=build_correlation_network_p,
    inputs={'correlation_table': PairwiseFeatureData},
    parameters={'max_val': Float % Range(0, 1, inclusive_end=True)},
    outputs=[('correlation_network', Network)],
    input_descriptions={'correlation_table': (
        'Pairwise feature data table of correlations with .')},
    parameter_descriptions={
        'max_val': 'The maximum p value to say an edge should exist.'
    },
    output_descriptions={'correlation_network': 'The resulting network.'},
    name='Build a correlation network based on a p value cutoff',
    description=(
        'Build a correlation network where nodes are features and edges '
        'are correlations with a p-value (or adjusted p-value) less than '
        'the max_val.'
    ),
)

# TODO: expose min_p based module making
plugin.methods.register_function(
    function=make_modules_on_correlations,
    inputs={'correlation_table': PairwiseFeatureData,
            'feature_table': FeatureTable[Frequency]},
    parameters={'min_r': Float % Range(0, 1, inclusive_end=True)},
    outputs=[('collapsed_table', FeatureTable[Frequency]),
             ('correlation_network', Network),
             ('module_membership', ModuleMembership)],
    input_descriptions={
        'correlation_table': 'Pairwise feature data table of correlations.',
        'feature_table': 'Feature table used to calculated correlations'
    },
    parameter_descriptions={'min_r': 'Minimum R value for creating an edge.'},
    output_descriptions={
        'collapsed_table': 'Feature table with module members collapsed into single observations.',
        'correlation_network': 'Network annotated with module membership and edges where correlation'
                               'was stronger than min_r',
        'module_membership': 'Feature metadata with module membership'
    },
    name='Find modules via the SCNIC method',
    description=(
        'Use the SCNIC method to find modules in a correlation network based '
        'on a minimum R value. A collapsed table and an annotated network are output.'
    ),
)

importlib.import_module('q2_SCNIC._transformer')



from biom.table import Table
import pandas as pd
from scipy.stats import spearmanr, pearsonr, kendalltau
import networkx as nx

from SCNIC.general import sparcc_paper_filter, correls_to_net, get_metadata_from_table, filter_correls
from SCNIC import correlation_analysis as ca
from SCNIC import module_analysis as ma


def sparcc_filter(table: Table) -> Table:
    return sparcc_paper_filter(table)


correl_methods = {'spearman': spearmanr, 'pearson': pearsonr, 'kendall': kendalltau, 'sparcc': 'sparcc'}


def calculate_correlations(table: Table, method: str, p_adjustment_method: str = 'fdr_bh', n_procs: int = 1) -> pd.DataFrame:
    print("Correlating with %s" % method)
    method = correl_methods[method]
    if method in [spearmanr, pearsonr, kendalltau]:
        correls = ca.calculate_correlations(table, method, p_adjustment_method=p_adjustment_method, nprocs=n_procs)
    elif method == 'sparcc':
        correls = ca.fastspar_correlation(table, verbose=True, nprocs=n_procs)
    else:
        raise ValueError('Provided correlation metric is not an accepted method.')
    return correls


def build_correlation_network_r(correlation_table: pd.DataFrame, min_val: float=.75,
                                cooccur: bool=False) -> nx.Graph:
    correlation_table_filtered = filter_correls(correlation_table, min_r=min_val, conet=cooccur)
    net = correls_to_net(correlation_table_filtered)
    return net


def build_correlation_network_p(correlation_table: pd.DataFrame, max_val: float=.05) -> nx.Graph:
    correlation_table_filtered = filter_correls(correlation_table, min_p=max_val)
    net = correls_to_net(correlation_table_filtered)
    return net


def make_modules_on_correlations(correlation_table: pd.DataFrame, feature_table: Table, min_r: float=.35) -> \
                                     (Table, nx.Graph, pd.Series):
    min_dist = ma.cor_to_dist(min_r)
    cor, labels = ma.correls_to_cor(correlation_table)
    dist = ma.cor_to_dist(cor)
    modules = ma.make_modules(dist, min_dist, obs_ids=labels)
    modules_rev = {asv: module for module, asvs in modules.items() for asv in asvs}
    for asv in feature_table.ids(axis='observation'):
        if asv not in modules_rev:
            modules_rev[asv] = None
    module_membership = pd.Series(modules_rev)
    coll_table = ma.collapse_modules(feature_table, modules)
    metadata = get_metadata_from_table(feature_table)
    metadata = ma.add_modules_to_metadata(modules, metadata)
    correlation_table_filtered = filter_correls(correlation_table, conet=True, min_r=min_r)
    net = correls_to_net(correlation_table_filtered, metadata=metadata)
    return coll_table, net, module_membership



from qiime2.plugin import SemanticType
from q2_types.feature_data import FeatureData

PairwiseFeatureData = SemanticType('PairwiseFeatureData')

Network = SemanticType('Network')

ModuleMembership = SemanticType('ModuleMembership', variant_of=FeatureData.field['type'])



from setuptools import find_packages, setup

__version__ = '2018.6.1'

setup(
    name='q2-SCNIC',
    version=__version__,
    license='BSD-3-Clause',
    packages=find_packages(),
    author="Michael Shaffer",
    author_email="michael.shaffer@ucdenver.edu",
    description=(
        "QIIME2 plugin for using SCNIC."),
    url="https://github.com/shafferm/q2-SCNIC",
    package_data={
        'q2_SCNIC': ['citations.bib']
    },
    entry_points={
        'qiime2.plugins':
        ['q2-SCNIC=q2_SCNIC.plugin_setup:plugin']
    }
)
